/*
 * RandomizerTestCase.java
 * 
 * Created 12.01.2006.
 * 
 * Copyright (c) 2006 Johann Burkard (<mailto:jb@eaio.com>)
 * <http://eaio.com>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
package com.eaio.util.junit;

import java.io.*;
import java.lang.reflect.*;
import java.security.SecureRandom;
import java.util.Random;

import junit.framework.TestCase;

/**
 * The RandomizerTestCase constructs instances of classes using random
 * parameters. It logs all method calls to a temporary file so that different
 * versions of a class may be compared.
 * 
 * @author <a href="mailto:jb@eaio.com">Johann Burkard</a>
 * @version $Id$
 */
public class RandomizerTestCase extends TestCase {

    /**
     * Empty serialized data.
     */
    private final byte[] SERIALIZED_DATA = new byte[] { (byte) 0xAC,
            (byte) 0xED, (byte) 0x00, (byte) 0x05 };

    /**
     * The cached test class.
     */
    private Class testClass;

    /**
     * The random number generator.
     */
    protected final Random rnd = new SecureRandom();

    private Object instance;

    /**
     * Constructor for RandomizerTestCase.
     */
    public RandomizerTestCase() {
        super();
    }

    /**
     * Constructor for RandomizerTestCase.
     *
     * @param name
     */
    public RandomizerTestCase(String name) {
        super(name);
    }

    public void testRandomized() {
        try {
            Object o = constructInstance(getTestClass());
            Class clazz = getTestClass();
            System.out.println(clazz.getName());
            Method[] methods = clazz.getMethods();
            for (int i = 0; i < methods.length; ++i) {
                System.out.print('#');
                System.out.println(methods[i].getName());
                System.out.println(callMethod(methods[i]));
            }

        }
        catch (Throwable thrw) {
            thrw.printStackTrace();
        }
    }

    /**
     * Returns the class to test. Should be overridden if the class name is not the
     * name of this class minus "Test".
     * 
     * @return the test class
     */
    protected Class getTestClass() {
        if (testClass == null) {
            String thisClazz = getClass().getName();
            String targetClassName = thisClazz.substring(0,
                    thisClazz.indexOf("Test"));
            try {
                testClass = Class.forName(targetClassName);
            }
            catch (ClassNotFoundException ex) {
                fail("Could not find test class '" + targetClassName
                        + "', please overwrite #getTestClass() if necessary.");
            }
            catch (RuntimeException ex) {
                fail("Could not find test class '" + targetClassName
                        + "', please overwrite #getTestClass() if necessary.");
            }
        }
        return testClass;
    }

    protected Object constructInstance(Class clazz) throws Throwable {
        Constructor[] constructors = clazz.getConstructors();
        Constructor c = null;
        if (constructors.length > 0 && rnd.nextBoolean()) {
            // call one of the public constructors
            c = constructors[rnd.nextInt(constructors.length)];
        }
        else {
            // call a non-public constructor
            constructors = clazz.getDeclaredConstructors();
            c = constructors[rnd.nextInt(constructors.length)];
            c.setAccessible(true);
        }
        Class[] parameterClasses = c.getParameterTypes();
        Object[] params = new Object[parameterClasses.length];
        for (int i = 0; i < params.length; ++i) {
            params[i] = randomObject(parameterClasses[i]);
        }
        instance = c.newInstance(params);
        return instance;
    }

    protected Object getInstance() {
        return instance;
    }

    /**
     * Returns the number of rounds methods should be called randomly.
     * <p>
     * This defaults to the number of methods times 8.
     * 
     * @return the number of test rounds
     */
    protected int getRunCount() {
        int size = getTestClass().getDeclaredMethods().length
                + getTestClass().getMethods().length;
        return size << 3;
    }

    /**
     * Returns the file used to store the results in or to read results from.
     * <p>
     * If the file does not exist, it will be created. If it does exist, the
     * contents will be read in and compared with the results of this run.
     * 
     * @return
     * @throws IOException
     */
    protected File getFile() throws IOException {
        String name = getTestClass().getName();
        File newTempFile = File.createTempFile(name, ".tmp");
        return newTempFile;
    }

    /**
     * Create an instance of the given type and optionally fill it with data.
     * 
     * @param type the type
     * @return an object or <code>null</code> if not found
     * @throws Exception
     */
    protected Object randomObject(Class type) throws Exception {
        if (type == File.class) {
            return File.createTempFile(getTestClass().getName(), ".tmp");
        }
        else if (ByteArrayInputStream.class == type) {
            return new ByteArrayInputStream((byte[]) randomObject(byte[].class));
        }
        else if (DataInput.class == type || DataInputStream.class == type) {
            return new DataInputStream(
                    (InputStream) randomObject(ByteArrayInputStream.class));
        }
        else if (ObjectInput.class == type || ObjectInputStream.class == type) {
            return new ObjectInputStream(new ByteArrayInputStream(
                    SERIALIZED_DATA));
        }
        else if (type == boolean[].class) {
            if (rnd.nextBoolean()) return null;
            boolean[] buf = new boolean[randomArrayLength()];
            if (rnd.nextBoolean()) {
                for (int i = 0; i < buf.length; ++i) {
                    buf[i] = rnd.nextBoolean();
                }
            }
        }
        else if (type == byte[].class) {
            if (rnd.nextBoolean()) return null;
            byte[] buf = new byte[randomArrayLength()];
            if (rnd.nextBoolean()) {
                for (int i = 0; i < buf.length; ++i) {
                    buf[i] = (byte) (rnd.nextInt(256) & 0xff);
                }
            }
            return buf;
        }
        else if (type == String.class) {
            if (rnd.nextBoolean()) return null;
            byte[] buf = null;
            while ((buf = (byte[]) randomObject(byte[].class)) == null);
            return new String(buf, 0);
        }
        else if (type.isInterface()) {
            if (rnd.nextBoolean()) return null;
            // Create a proxy
            return Proxy.newProxyInstance(type.getClassLoader(),
                    new Class[] { type }, new RandomizerInvocationHandler());
        }
        return null;
    }

    protected int randomArrayLength() {
        if (rnd.nextBoolean()) {
            return 0;
        }
        else {
            // Large numbers are ok, but insanely large numbers are not.
            return rnd.nextInt(424242);
        }
    }

    protected Object[] randomObjects(Method method) throws Exception {
        Class[] parameters = method.getParameterTypes();
        if (parameters == null || parameters.length == 0) {
            return null;
        }
        Object[] out = new Object[parameters.length];
        for (int i = 0; i < out.length; ++i) {
            out[i] = randomObject(parameters[i]);
        }
        return out;
    }

    protected Object callMethod(Method method) throws Throwable {
        try {
            return method.invoke(instance, randomObjects(method));
        }
        catch (InvocationTargetException ex) {
            System.out.println(method);
            throw ((InvocationTargetException) ex).getTargetException();
        }
    }

    protected class RandomizerInvocationHandler implements InvocationHandler {

        /**
         * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
         */
        public Object invoke(Object proxy, Method method, Object[] args)
                throws Throwable {
            return callMethod(method);
        }

    }

}
